# Copyright 2017 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading

import grpc
from grpc_testing import _common


class State(_common.ChannelRpcHandler):

    def __init__(self, invocation_metadata, requests, requests_closed):
        self._condition = threading.Condition()
        self._invocation_metadata = invocation_metadata
        self._requests = requests
        self._requests_closed = requests_closed
        self._initial_metadata = None
        self._responses = []
        self._trailing_metadata = None
        self._code = None
        self._details = None

    def initial_metadata(self):
        with self._condition:
            while True:
                if self._initial_metadata is None:
                    if self._code is None:
                        self._condition.wait()
                    else:
                        return _common.FUSSED_EMPTY_METADATA
                else:
                    return self._initial_metadata

    def add_request(self, request):
        with self._condition:
            if self._code is None and not self._requests_closed:
                self._requests.append(request)
                self._condition.notify_all()
                return True
            else:
                return False

    def close_requests(self):
        with self._condition:
            if self._code is None and not self._requests_closed:
                self._requests_closed = True
                self._condition.notify_all()

    def take_response(self):
        with self._condition:
            while True:
                if self._code is grpc.StatusCode.OK:
                    if self._responses:
                        response = self._responses.pop(0)
                        return _common.ChannelRpcRead(response, None, None,
                                                      None)
                    else:
                        return _common.ChannelRpcRead(None,
                                                      self._trailing_metadata,
                                                      grpc.StatusCode.OK,
                                                      self._details)
                elif self._code is None:
                    if self._responses:
                        response = self._responses.pop(0)
                        return _common.ChannelRpcRead(response, None, None,
                                                      None)
                    else:
                        self._condition.wait()
                else:
                    return _common.ChannelRpcRead(None, self._trailing_metadata,
                                                  self._code, self._details)

    def termination(self):
        with self._condition:
            while True:
                if self._code is None:
                    self._condition.wait()
                else:
                    return self._trailing_metadata, self._code, self._details

    def cancel(self, code, details):
        with self._condition:
            if self._code is None:
                if self._initial_metadata is None:
                    self._initial_metadata = _common.FUSSED_EMPTY_METADATA
                self._trailing_metadata = _common.FUSSED_EMPTY_METADATA
                self._code = code
                self._details = details
                self._condition.notify_all()
                return True
            else:
                return False

    def take_invocation_metadata(self):
        with self._condition:
            if self._invocation_metadata is None:
                raise ValueError('Expected invocation metadata!')
            else:
                invocation_metadata = self._invocation_metadata
                self._invocation_metadata = None
                return invocation_metadata

    def take_invocation_metadata_and_request(self):
        with self._condition:
            if self._invocation_metadata is None:
                raise ValueError('Expected invocation metadata!')
            elif not self._requests:
                raise ValueError('Expected at least one request!')
            else:
                invocation_metadata = self._invocation_metadata
                self._invocation_metadata = None
                return invocation_metadata, self._requests.pop(0)

    def send_initial_metadata(self, initial_metadata):
        with self._condition:
            self._initial_metadata = _common.fuss_with_metadata(
                initial_metadata)
            self._condition.notify_all()

    def take_request(self):
        with self._condition:
            while True:
                if self._requests:
                    return self._requests.pop(0)
                else:
                    self._condition.wait()

    def requests_closed(self):
        with self._condition:
            while True:
                if self._requests_closed:
                    return
                else:
                    self._condition.wait()

    def send_response(self, response):
        with self._condition:
            if self._code is None:
                self._responses.append(response)
                self._condition.notify_all()

    def terminate_with_response(self, response, trailing_metadata, code,
                                details):
        with self._condition:
            if self._initial_metadata is None:
                self._initial_metadata = _common.FUSSED_EMPTY_METADATA
            self._responses.append(response)
            self._trailing_metadata = _common.fuss_with_metadata(
                trailing_metadata)
            self._code = code
            self._details = details
            self._condition.notify_all()

    def terminate(self, trailing_metadata, code, details):
        with self._condition:
            if self._initial_metadata is None:
                self._initial_metadata = _common.FUSSED_EMPTY_METADATA
            self._trailing_metadata = _common.fuss_with_metadata(
                trailing_metadata)
            self._code = code
            self._details = details
            self._condition.notify_all()

    def cancelled(self):
        with self._condition:
            while True:
                if self._code is grpc.StatusCode.CANCELLED:
                    return
                elif self._code is None:
                    self._condition.wait()
                else:
                    raise ValueError('Status code unexpectedly {}!'.format(
                        self._code))

    def is_active(self):
        raise NotImplementedError()

    def time_remaining(self):
        raise NotImplementedError()

    def add_callback(self, callback):
        raise NotImplementedError()



# Copyright 2017 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

import grpc


class _RequestIterator(object):

    def __init__(self, rpc, handler):
        self._rpc = rpc
        self._handler = handler

    def _next(self):
        read = self._handler.take_request()
        if read.requests_closed:
            raise StopIteration()
        elif read.terminated:
            rpc_error = grpc.RpcError()
            self._rpc.add_rpc_error(rpc_error)
            raise rpc_error
        else:
            return read.request

    def __iter__(self):
        return self

    def __next__(self):
        return self._next()

    def next(self):
        return self._next()


def _unary_response(argument, implementation, rpc, servicer_context):
    try:
        response = implementation(argument, servicer_context)
    except Exception as exception:  # pylint: disable=broad-except
        rpc.application_exception_abort(exception)
    else:
        rpc.unary_response_complete(response)


def _stream_response(argument, implementation, rpc, servicer_context):
    try:
        response_iterator = implementation(argument, servicer_context)
    except Exception as exception:  # pylint: disable=broad-except
        rpc.application_exception_abort(exception)
    else:
        while True:
            try:
                response = copy.deepcopy(next(response_iterator))
            except StopIteration:
                rpc.stream_response_complete()
                break
            except Exception as exception:  # pylint: disable=broad-except
                rpc.application_exception_abort(exception)
                break
            else:
                rpc.stream_response(response)


def unary_unary(implementation, rpc, request, servicer_context):
    _unary_response(request, implementation, rpc, servicer_context)


def unary_stream(implementation, rpc, request, servicer_context):
    _stream_response(request, implementation, rpc, servicer_context)


def stream_unary(implementation, rpc, handler, servicer_context):
    _unary_response(_RequestIterator(rpc, handler), implementation, rpc,
                    servicer_context)


def stream_stream(implementation, rpc, handler, servicer_context):
    _stream_response(_RequestIterator(rpc, handler), implementation, rpc,
                     servicer_context)



# Copyright 2017 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import threading

import grpc
from grpc_testing import _common

logging.basicConfig()
_LOGGER = logging.getLogger(__name__)


class Rpc(object):

    def __init__(self, handler, invocation_metadata):
        self._condition = threading.Condition()
        self._handler = handler
        self._invocation_metadata = invocation_metadata
        self._initial_metadata_sent = False
        self._pending_trailing_metadata = None
        self._pending_code = None
        self._pending_details = None
        self._callbacks = []
        self._active = True
        self._rpc_errors = []

    def _ensure_initial_metadata_sent(self):
        if not self._initial_metadata_sent:
            self._handler.send_initial_metadata(_common.FUSSED_EMPTY_METADATA)
            self._initial_metadata_sent = True

    def _call_back(self):
        callbacks = tuple(self._callbacks)
        self._callbacks = None

        def call_back():
            for callback in callbacks:
                try:
                    callback()
                except Exception:  # pylint: disable=broad-except
                    _LOGGER.exception('Exception calling server-side callback!')

        callback_calling_thread = threading.Thread(target=call_back)
        callback_calling_thread.start()

    def _terminate(self, trailing_metadata, code, details):
        if self._active:
            self._active = False
            self._handler.send_termination(trailing_metadata, code, details)
            self._call_back()
            self._condition.notify_all()

    def _complete(self):
        if self._pending_trailing_metadata is None:
            trailing_metadata = _common.FUSSED_EMPTY_METADATA
        else:
            trailing_metadata = self._pending_trailing_metadata
        if self._pending_code is None:
            code = grpc.StatusCode.OK
        else:
            code = self._pending_code
        details = '' if self._pending_details is None else self._pending_details
        self._terminate(trailing_metadata, code, details)

    def _abort(self, code, details):
        self._terminate(_common.FUSSED_EMPTY_METADATA, code, details)

    def add_rpc_error(self, rpc_error):
        with self._condition:
            self._rpc_errors.append(rpc_error)

    def application_cancel(self):
        with self._condition:
            self._abort(grpc.StatusCode.CANCELLED,
                        'Cancelled by server-side application!')

    def application_exception_abort(self, exception):
        with self._condition:
            if exception not in self._rpc_errors:
                _LOGGER.exception('Exception calling application!')
                self._abort(
                    grpc.StatusCode.UNKNOWN,
                    'Exception calling application: {}'.format(exception))

    def extrinsic_abort(self):
        with self._condition:
            if self._active:
                self._active = False
                self._call_back()
                self._condition.notify_all()

    def unary_response_complete(self, response):
        with self._condition:
            self._ensure_initial_metadata_sent()
            self._handler.add_response(response)
            self._complete()

    def stream_response(self, response):
        with self._condition:
            self._ensure_initial_metadata_sent()
            self._handler.add_response(response)

    def stream_response_complete(self):
        with self._condition:
            self._ensure_initial_metadata_sent()
            self._complete()

    def send_initial_metadata(self, initial_metadata):
        with self._condition:
            if self._initial_metadata_sent:
                return False
            else:
                self._handler.send_initial_metadata(initial_metadata)
                self._initial_metadata_sent = True
                return True

    def is_active(self):
        with self._condition:
            return self._active

    def add_callback(self, callback):
        with self._condition:
            if self._callbacks is None:
                return False
            else:
                self._callbacks.append(callback)
                return True

    def invocation_metadata(self):
        with self._condition:
            return self._invocation_metadata

    def set_trailing_metadata(self, trailing_metadata):
        with self._condition:
            self._pending_trailing_metadata = trailing_metadata

    def set_code(self, code):
        with self._condition:
            self._pending_code = code

    def set_details(self, details):
        with self._condition:
            self._pending_details = details



# Copyright 2017 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading

import grpc_testing
from grpc_testing import _common
from grpc_testing._server import _handler
from grpc_testing._server import _rpc
from grpc_testing._server import _server_rpc
from grpc_testing._server import _service
from grpc_testing._server import _servicer_context


def _implementation(descriptors_to_servicers, method_descriptor):
    servicer = descriptors_to_servicers[method_descriptor.containing_service]
    return getattr(servicer, method_descriptor.name)


def _unary_unary_service(request):

    def service(implementation, rpc, servicer_context):
        _service.unary_unary(implementation, rpc, request, servicer_context)

    return service


def _unary_stream_service(request):

    def service(implementation, rpc, servicer_context):
        _service.unary_stream(implementation, rpc, request, servicer_context)

    return service


def _stream_unary_service(handler):

    def service(implementation, rpc, servicer_context):
        _service.stream_unary(implementation, rpc, handler, servicer_context)

    return service


def _stream_stream_service(handler):

    def service(implementation, rpc, servicer_context):
        _service.stream_stream(implementation, rpc, handler, servicer_context)

    return service


class _Serverish(_common.Serverish):

    def __init__(self, descriptors_to_servicers, time):
        self._descriptors_to_servicers = descriptors_to_servicers
        self._time = time

    def _invoke(self, service_behavior, method_descriptor, handler,
                invocation_metadata, deadline):
        implementation = _implementation(self._descriptors_to_servicers,
                                         method_descriptor)
        rpc = _rpc.Rpc(handler, invocation_metadata)
        if handler.add_termination_callback(rpc.extrinsic_abort):
            servicer_context = _servicer_context.ServicerContext(
                rpc, self._time, deadline)
            service_thread = threading.Thread(target=service_behavior,
                                              args=(
                                                  implementation,
                                                  rpc,
                                                  servicer_context,
                                              ))
            service_thread.start()

    def invoke_unary_unary(self, method_descriptor, handler,
                           invocation_metadata, request, deadline):
        self._invoke(_unary_unary_service(request), method_descriptor, handler,
                     invocation_metadata, deadline)

    def invoke_unary_stream(self, method_descriptor, handler,
                            invocation_metadata, request, deadline):
        self._invoke(_unary_stream_service(request), method_descriptor, handler,
                     invocation_metadata, deadline)

    def invoke_stream_unary(self, method_descriptor, handler,
                            invocation_metadata, deadline):
        self._invoke(_stream_unary_service(handler), method_descriptor, handler,
                     invocation_metadata, deadline)

    def invoke_stream_stream(self, method_descriptor, handler,
                             invocation_metadata, deadline):
        self._invoke(_stream_stream_service(handler), method_descriptor,
                     handler, invocation_metadata, deadline)


def _deadline_and_handler(requests_closed, time, timeout):
    if timeout is None:
        return None, _handler.handler_without_deadline(requests_closed)
    else:
        deadline = time.time() + timeout
        handler = _handler.handler_with_deadline(requests_closed, time,
                                                 deadline)
        return deadline, handler


class _Server(grpc_testing.Server):

    def __init__(self, serverish, time):
        self._serverish = serverish
        self._time = time

    def invoke_unary_unary(self, method_descriptor, invocation_metadata,
                           request, timeout):
        deadline, handler = _deadline_and_handler(True, self._time, timeout)
        self._serverish.invoke_unary_unary(method_descriptor, handler,
                                           invocation_metadata, request,
                                           deadline)
        return _server_rpc.UnaryUnaryServerRpc(handler)

    def invoke_unary_stream(self, method_descriptor, invocation_metadata,
                            request, timeout):
        deadline, handler = _deadline_and_handler(True, self._time, timeout)
        self._serverish.invoke_unary_stream(method_descriptor, handler,
                                            invocation_metadata, request,
                                            deadline)
        return _server_rpc.UnaryStreamServerRpc(handler)

    def invoke_stream_unary(self, method_descriptor, invocation_metadata,
                            timeout):
        deadline, handler = _deadline_and_handler(False, self._time, timeout)
        self._serverish.invoke_stream_unary(method_descriptor, handler,
                                            invocation_metadata, deadline)
        return _server_rpc.StreamUnaryServerRpc(handler)

    def invoke_stream_stream(self, method_descriptor, invocation_metadata,
                             timeout):
        deadline, handler = _deadline_and_handler(False, self._time, timeout)
        self._serverish.invoke_stream_stream(method_descriptor, handler,
                                             invocation_metadata, deadline)
        return _server_rpc.StreamStreamServerRpc(handler)


def server_from_descriptor_to_servicers(descriptors_to_servicers, time):
    return _Server(_Serverish(descriptors_to_servicers, time), time)
